@@ -39,8 +39,9 @@ describe Rack::Session::Redis do
 
   it "can create it's own pool" do
     session_store = Rack::Session::Redis.new(incrementor, pool_size: 5, pool_timeout: 10)
-    session_store.pool.class.must_equal ::ConnectionPool
-    session_store.pool.instance_variable_get(:@size).must_equal 5
+    conn = session_store.instance_variable_get(:@conn)
+    conn.pool.class.must_equal ::ConnectionPool
+    conn.pool.instance_variable_get(:@size).must_equal 5
   end
 
   it "can create it's own pool using default Redis server" do
@@ -53,33 +54,77 @@ describe Rack::Session::Redis do
     session_store.with { |connection| connection.to_s.must_match(/127\.0\.0\.1:6380 against DB 1$/)  }
   end
 
-
   it "can use a supplied pool" do
     session_store = Rack::Session::Redis.new(incrementor, pool: ::ConnectionPool.new(size: 1, timeout: 1) { ::Redis::Store::Factory.create("redis://127.0.0.1:6380/1")})
-    session_store.pool.class.must_equal ::ConnectionPool
-    session_store.pool.instance_variable_get(:@size).must_equal 1
+    conn = session_store.instance_variable_get(:@conn)
+    conn.pool.class.must_equal ::ConnectionPool
+    conn.pool.instance_variable_get(:@size).must_equal 1
   end
 
   it "uses the specified Redis store when provided" do
     store = ::Redis::Store::Factory.create('redis://127.0.0.1:6380/1')
     pool = Rack::Session::Redis.new(incrementor, :redis_store => store)
-    pool.pool.to_s.must_match(/127\.0\.0\.1:6380 against DB 1$/)
-    pool.pool.must_equal(store)
+    pool.with do |p|
+      p.to_s.must_match(/127\.0\.0\.1:6380 against DB 1$/)
+      p.must_equal(store)
+    end
   end
 
   it "uses the default Redis server and namespace when not provided" do
     pool = Rack::Session::Redis.new(incrementor)
-    pool.pool.to_s.must_match(/127\.0\.0\.1:6379 against DB 0 with namespace rack:session$/)
+    pool.with { |p| p.to_s.must_match(/127\.0\.0\.1:6379 against DB 0 with namespace rack:session$/) }
   end
 
   it "uses the specified namespace when provided" do
     pool = Rack::Session::Redis.new(incrementor, :redis_server => {:namespace => 'test:rack:session'})
-    pool.pool.to_s.must_match(/namespace test:rack:session$/)
+    pool.with { |p| p.to_s.must_match(/namespace test:rack:session$/) }
   end
 
   it "uses the specified Redis server when provided" do
     pool = Rack::Session::Redis.new(incrementor, :redis_server => 'redis://127.0.0.1:6380/1')
-    pool.pool.to_s.must_match(/127\.0\.0\.1:6380 against DB 1$/)
+    pool.with { |p| p.to_s.must_match(/127\.0\.0\.1:6380 against DB 1$/) }
+  end
+
+  it "is threadsafe by default" do
+    sesion_store = Rack::Session::Redis.new(incrementor)
+    sesion_store.threadsafe?.must_equal(true)
+  end
+
+  it "does not store a blank session" do
+    session_store = Rack::Session::Redis.new(incrementor)
+    sid = session_store.generate_unique_sid({})
+    session_store.with { |c| c.get(sid.private_id).must_be_nil }
+  end
+
+  it "locks the store mutex" do
+    mutex = Mutex.new
+    mutex.expects(:lock).once
+    sesion_store = Rack::Session::Redis.new(incrementor)
+    sesion_store.instance_variable_set(:@mutex, mutex)
+    was_yielded = false
+    request = Minitest::Mock.new
+    request.expect(:multithread?, true)
+    sesion_store.with_lock(request) { was_yielded = true}
+    was_yielded.must_equal(true)
+  end
+
+  describe "threadsafe disabled" do
+    it "can have the global lock disabled" do
+      sesion_store = Rack::Session::Redis.new(incrementor, :threadsafe => false)
+      sesion_store.threadsafe?.must_equal(false)
+    end
+
+    it "does not lock the store mutex" do
+      mutex = Mutex.new
+      mutex.expects(:lock).never
+      sesion_store = Rack::Session::Redis.new(incrementor, :threadsafe => false)
+      sesion_store.instance_variable_set(:@mutex, mutex)
+      was_yielded = false
+      request = Minitest::Mock.new
+      request.expect(:multithread?, true)
+      sesion_store.with_lock(request) { was_yielded = true}
+      was_yielded.must_equal(true)
+    end
   end
 
   it "creates a new cookie" do
@@ -229,15 +274,12 @@ describe Rack::Session::Redis do
   end
 
   it "does not hit with :skip option" do
-    with_pool_management(incrementor) do |pool|
-      skip = Rack::Utils::Context.new(pool, skip_session)
+    with_pool_management(incrementor) do |session_store|
+      skip = Rack::Utils::Context.new(session_store, skip_session)
       sreq = Rack::MockRequest.new(skip)
 
-      pool.instance_variable_set('@pool', MiniTest::Mock.new)
-
       res0 = sreq.get("/")
       res0.body.must_equal('{"counter"=>1}')
-      assert pool.pool.verify
     end
   end
 
@@ -257,10 +299,11 @@ describe Rack::Session::Redis do
 
       res0 = req.get("/")
       session_id = (cookie = res0["Set-Cookie"])[session_match, 1]
-      ses0 = pool.with { |c| c.get(session_id) }
+      sid = Rack::Session::SessionId.new(session_id)
+      ses0 = pool.with { |c| c.get(sid.private_id) }
 
       req.get("/", "HTTP_COOKIE" => cookie)
-      ses1 = pool.with { |c| c.get(session_id) }
+      ses1 = pool.with { |c| c.get(sid.private_id) }
 
       ses1.wont_equal(ses0)
     end
@@ -280,6 +323,7 @@ describe Rack::Session::Redis do
       res.body.must_equal('{"counter"=>1}')
       cookie = res["Set-Cookie"]
       session_id = cookie[session_match, 1]
+      sid = Rack::Session::SessionId.new(session_id)
 
       delta_incrementor = lambda do |env|
         # emulate disconjoinment of threading
@@ -301,7 +345,7 @@ describe Rack::Session::Redis do
         request.body.must_include('"counter"=>2')
       end
 
-      session = pool.with { |c| c.get(session_id) }
+      session = pool.with { |c| c.get(sid.private_id) }
       session.size.must_equal(tnum+1) # counter
       session['counter'].must_equal(2) # meeeh
 
@@ -318,7 +362,7 @@ describe Rack::Session::Redis do
         request.body.must_include('"counter"=>3')
       end
 
-      session = pool.with { |c| c.get(session_id) }
+      session = pool.with { |c| c.get(sid.private_id) }
       session.size.must_equal(tnum+1)
       session['counter'].must_equal(3)
 
@@ -340,7 +384,7 @@ describe Rack::Session::Redis do
         request.body.must_include('"foo"=>"bar"')
       end
 
-      session = pool.with { |c| c.get(session_id) }
+      session = pool.with { |c| c.get(sid.private_id) }
       session.size.must_equal(r.size+1)
       session['counter'].must_be_nil
       session['foo'].must_equal('bar')
